# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See http://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------


from ctypes import Array, create_string_buffer, c_double, c_float
from numpy import asarray, asmatrix, ascontiguousarray, ctypeslib, float64, ndarray
import re
from sys import version_info
from .utils import is_pypy


def __mac_address_from_bytes(a, b, c, d, e, f):
  """
  Used internally by ``to_mac_address``
  """
  from .lookup import MacAddress
  return MacAddress(int(a), int(b), int(c), int(d), int(e), int(f))


def __mac_address_from_string(address):
  """
  Used internally by `to_mac_address`
  """
  from .lookup import MacAddress
  if type(address) != str:
    try:
      address = str(address)
    except:
      raise ValueError('Input must be string or convertible to string')

  match = re.match(r'^(([a-fA-F0-9]{2}):){5}([a-fA-F0-9]{2})$', address)
  if match != None:
    mac_byte_a = int(address[0:2], 16)
    mac_byte_b = int(address[3:5], 16)
    mac_byte_c = int(address[6:8], 16)
    mac_byte_d = int(address[9:11], 16)
    mac_byte_e = int(address[12:14], 16)
    mac_byte_f = int(address[15:17], 16)
    return MacAddress(mac_byte_a, mac_byte_b, mac_byte_c,
                      mac_byte_d, mac_byte_e, mac_byte_f)
  else:
    raise ValueError('Unable to parse mac address'
                     ' from string {0}'.format(address))


def to_mac_address(*args):
  """
  Convert input argument(s) to a MacAddress object.
  Only 1 or 6 arguments are valid.

  If 1 argument is provided, try the following:

    * If input type is MacAddress, simply return that object
    * If input type is list or ctypes Array, recall with these elements
    * If input is of another type, try to parse a MAC address from its
      `__str__` representation

  When 6 parameters are provided, this attempts to construct a MAC address
  by interpreting the input parameters as sequential bytes of a mac address.

  If the provided argument count is neither 1 or 6,
  this function throws an exception.

  :param args: 1 or 6 element list of variadic arguments
  :return: a MacAddress instance
  """
  from .lookup import MacAddress

  if len(args) == 1:
    if type(args[0]) is MacAddress:
      return args[0]
    elif isinstance(args[0], list) or isinstance(args[0], Array):
      if len(args[0]) == 1:
        return to_mac_address(args[0])
      elif len(args[0]) == 6:
        arg = args[0]
        return to_mac_address(*arg)
      else:
        raise ValueError('Invalid amount of arguments provided'
                         ' ({0}). Expected 1 or 6'.format(len(args[0])))
    else:
      try:
        return __mac_address_from_string(args[0])
      except ValueError as v:
        raise ValueError('Could not create mac address from argument', v)
  elif len(args) == 6:
    return __mac_address_from_bytes(*args)
  else:
    raise ValueError('Invalid amount of arguments provided'
                     ' ({0}). Expected 1 or 6'.format(len(args)))


################################################################################
# Converting to numpy types
################################################################################


def __to_contig_sq_mat_chk_size(size):
  try:
    size = int(size)
  except Exception as e:
    raise ValueError('size must be convertible to an integer', e)
  if size < 1:
    raise ValueError('size must be greater than zero')
  return size


def __to_contig_sq_mat_handle_ret(ret, size):
  expected_shape = (size, size)
  if ret.shape != expected_shape:
    ret = ret.reshape(expected_shape)

  # Enforce output will be right shape
  if ret.shape != expected_shape:
    raise ValueError('Cannot convert input to shape {}'.format(expected_shape))

  # Enforce contiguous in memory
  if not ret.flags['C_CONTIGUOUS']:
    ret = ascontiguousarray(ret)
  return ret


if is_pypy():

  __pypy_ctypes_arr_types = dict()

  def __to_contig_sq_mat_impl(mat, dtype, size):
    size = __to_contig_sq_mat_chk_size(size)
    if isinstance(mat, str):
      # Numpy does not let you call np.asarray(str, dtype=float_type) with a string input, but `np.asmatrix` does.
      # This seems like a hack and should be removed in the near future.
      ret = asmatrix(mat, dtype=dtype)
    else:
      # Hack to fix weird pypy numpy issues with __buffer__
      if isinstance(mat, Array) and not hasattr(mat, '_HEBI_PYPY_HACK'):
        mat = mat[:]
      ret = asarray(mat, dtype=dtype)
    return __to_contig_sq_mat_handle_ret(ret, size)

else:

  def __to_contig_sq_mat_impl(mat, dtype, size):
    size = __to_contig_sq_mat_chk_size(size)
    if isinstance(mat, str):
      # Numpy does not let you call np.asarray(str, dtype=float_type) with a string input, but `np.asmatrix` does.
      # This seems like a hack and should be removed in the near future.
      ret = asmatrix(mat, dtype=dtype)
    else:
      ret = asarray(mat, dtype=dtype)
    return __to_contig_sq_mat_handle_ret(ret, size)


def to_contig_sq_mat(mat, dtype=float64, size=3):
  """
  Converts input to a numpy square matrix of the specified data type and size.

  This function ensures that the underlying data is laid out
  in contiguous memory.

  :param mat: Input matrix
  :param dtype: Data type of matrix
  :param size: Size of matrix
  :return: a `size`x`size` numpy matrix with elements of type `dtype`
  """
  return __to_contig_sq_mat_impl(mat, dtype, size)


if is_pypy():

  from ctypes import POINTER
  _CTypes_LP_Double = POINTER(c_double)

  class _PyPyLP_Double(_CTypes_LP_Double):
    _type_ = _CTypes_LP_Double._type_
    __slots__ = ['_HEBI_PYPY_HACK']

    def __init__(self, val):
      super(_PyPyLP_Double, self).__init__(val)
      self._HEBI_PYPY_HACK = True
    def __buffer__(self, *args):
      pass


  def np_array_from_dbl_ptr(ptr, size):
    return ctypeslib.as_array(_PyPyLP_Double(ptr.contents), size)

else:

  def np_array_from_dbl_ptr(ptr, size):
    return ctypeslib.as_array(ptr, size)


################################################################################
# CTypes Compatibility functions
################################################################################

if version_info[0] == 2:

  create_string_buffer_compat = create_string_buffer

  def decode_string_buffer(bfr, encoding='utf8'):
    import ctypes
    if type(bfr) == str:
      return bfr
    elif isinstance(bfr, ctypes.Array):
      return ctypes.cast(bfr, ctypes.c_char_p).value
    else:
      raise TypeError(bfr)


  def __is_int_type(val):
    return isinstance(val, (int, long))


else:

  def create_string_buffer_compat(init, size=None):
    if size is None:
      if isinstance(init, str):
        return create_string_buffer(bytes(init, 'utf8'))
      elif isinstance(init, int):
        return create_string_buffer(init)
      raise TypeError(init)
    else:
      return create_string_buffer(bytes(init, 'utf8'), size)


  def decode_string_buffer(bfr, encoding='utf8'):
    """
    Enables compatibility between Python 2 and 3

    :param bfr: a string, ``bytes``, or ctypes array
    :return: a string
    """
    import ctypes
    if type(bfr) == str:
      return bfr
    elif isinstance(bfr, bytes):
      return bfr.decode(encoding, 'replace')
    elif isinstance(bfr, ctypes.Array):
      casted = ctypes.cast(bfr, ctypes.c_char_p).value
      if casted is None:
        return None
      return casted.decode(encoding, 'replace')
    else:
      raise TypeError(bfr)


  def __is_int_type(val):
    return isinstance(val, int)


def __dbl_bf_check(size):
  if not __is_int_type(size):
    raise TypeError('size must be an integer')
  if size < 1:
    raise ValueError('size must be a positive number')


if is_pypy():
  # Work around some ridiculous PyPy numpy-ctypes bugs here

  __pypy_dbl_bfr_types = dict()
  __pypy_float_bfr_types = dict()

  def __terrible_Ctypes_Array_Type_Hack(cls):
    class PyPyBufferType(cls):
      __slots__ = ['_HEBI_PYPY_HACK']
      def __init__(self, *args):
        super(PyPyBufferType, self).__init__(*args)
        self._HEBI_PYPY_HACK = True

      def __buffer__(self, *args, **kwargs):
        # For some reason, numpy on PyPy wants to call buffer,
        # which will return a __builtin__.buffer object, which
        # is absolutely wrong.
        # If numpy succeeds in calling buffer(), the dtype will become `uint8`,
        # which is obviously entirely wrong here.
        pass
    return PyPyBufferType


  def create_float_buffer(size):
    __dbl_bf_check(size)
    try:
      PyPyBufferType = __pypy_float_bfr_types[size]
    except KeyError:
      PyPyBufferType = __terrible_Ctypes_Array_Type_Hack((c_float*size))
      __pypy_float_bfr_types[size] = PyPyBufferType
    return PyPyBufferType()


  # This works around very strange pypy numpy bugs
  def create_double_buffer(size):
    __dbl_bf_check(size)
    try:
      PyPyBufferType = __pypy_dbl_bfr_types[size]
    except KeyError:
      PyPyBufferType = __terrible_Ctypes_Array_Type_Hack((c_double*size))
      __pypy_dbl_bfr_types[size] = PyPyBufferType
    return PyPyBufferType()


  # Cache used sizes
  create_double_buffer(6)
  create_double_buffer(16)
  create_double_buffer(256)

  create_float_buffer(3)
  create_float_buffer(4)

else:

  __dbl_bfr_types = dict()
  __float_bfr_types = dict()

  def create_double_buffer(size):
    """
    Creates a ctypes array of c_double elements

    :param size: The number of elements to be in the array
    :return: c_double array
    """
    __dbl_bf_check(size)
    try:
      ArrType = __dbl_bfr_types[size]
    except KeyError:
      ArrType = (c_double*size)
      __dbl_bfr_types[size] = ArrType
    return ArrType()

  def create_float_buffer(size):
    """
    Creates a ctypes array of c_float elements

    :param size: The number of elements to be in the array
    :return: c_float array
    """
    __dbl_bf_check(size)
    try:
      ArrType = __float_bfr_types[size]
    except KeyError:
      ArrType = (c_float*size)
      __float_bfr_types[size] = ArrType
    return ArrType()

  # Cache used sizes
  __dbl_bfr_types[6] = (c_double*6)
  __dbl_bfr_types[16] = (c_double*16)
  __dbl_bfr_types[256] = (c_double*256)

  __float_bfr_types[3] = (c_float*3)
  __float_bfr_types[4] = (c_float*4)



from .wrappers import EnumTraits, EnumType

# Status Codes
StatusSuccess = EnumTraits(0, 'StatusSuccess')
StatusInvalidArgument = EnumTraits(1, 'StatusInvalidArgument')
StatusBufferTooSmall = EnumTraits(2, 'StatusBufferTooSmall')
StatusValueNotSet = EnumTraits(3, 'StatusValueNotSet')
StatusFailure = EnumTraits(4, 'StatusFailure')
StatusArgumentOutOfRange = EnumTraits(5, 'StatusArgumentOutOfRange')
StatusCode = EnumType('StatusCode', [StatusSuccess, StatusInvalidArgument, StatusBufferTooSmall, StatusValueNotSet, StatusFailure, StatusArgumentOutOfRange])

# Frame Types
FrameTypeCenterOfMass = EnumTraits(0, 'FrameTypeCenterOfMass')
FrameTypeOutput = EnumTraits(1, 'FrameTypeOutput')
FrameTypeEndEffector = EnumTraits(2, 'FrameTypeEndEffector')
FrameTypeInput = EnumTraits(3, 'FrameTypeInput')
FrameType = EnumType('FrameType', [FrameTypeCenterOfMass, FrameTypeOutput, FrameTypeEndEffector, FrameTypeInput])

# Robot Model meta data
RobotModelElementTypeOther = EnumTraits(0, 'RobotModelElementTypeOther')
RobotModelElementTypeActuator = EnumTraits(1, 'RobotModelElementTypeActuator')
RobotModelElementTypeBracket = EnumTraits(2, 'RobotModelElementTypeBracket')
RobotModelElementTypeJoint = EnumTraits(3, 'RobotModelElementTypeJoint')
RobotModelElementTypeLink = EnumTraits(4, 'RobotModelElementTypeLink')
RobotModelElementTypeRigidBody = EnumTraits(5, 'RobotModelElementTypeRigidBody')
RobotModelElementTypeEndEffector = EnumTraits(6, 'RobotModelElementTypeEndEffector')
RobotModelElementType = EnumType('RobotModelElementType', [RobotModelElementTypeOther, RobotModelElementTypeActuator, RobotModelElementTypeBracket, RobotModelElementTypeJoint, RobotModelElementTypeLink, RobotModelElementTypeRigidBody, RobotModelElementTypeEndEffector])

# Types of motion allowed by joints
JointTypeRotationX = EnumTraits(0, 'JointTypeRotationX')
JointTypeRotationY = EnumTraits(1, 'JointTypeRotationY')
JointTypeRotationZ = EnumTraits(2, 'JointTypeRotationZ')
JointTypeTranslationX = EnumTraits(3, 'JointTypeTranslationX')
JointTypeTranslationY = EnumTraits(4, 'JointTypeTranslationY')
JointTypeTranslationZ = EnumTraits(5, 'JointTypeTranslationZ')
JointType = EnumType('JointType', [JointTypeRotationX, JointTypeRotationY, JointTypeRotationZ, JointTypeTranslationX, JointTypeTranslationY, JointTypeTranslationZ])

# Actuator Types
ActuatorTypeX5_1 = EnumTraits(0, 'ActuatorTypeX5_1')
ActuatorTypeX5_4 = EnumTraits(1, 'ActuatorTypeX5_4')
ActuatorTypeX5_9 = EnumTraits(2, 'ActuatorTypeX5_9')
ActuatorTypeX8_3 = EnumTraits(3, 'ActuatorTypeX8_3')
ActuatorTypeX8_9 = EnumTraits(4, 'ActuatorTypeX8_9')
ActuatorTypeX8_16 = EnumTraits(5, 'ActuatorTypeX8_16')
ActuatorTypeR8_3 = EnumTraits(6, 'ActuatorTypeR8_16')
ActuatorTypeR8_9 = EnumTraits(7, 'ActuatorTypeR8_16')
ActuatorTypeR8_16 = EnumTraits(8, 'ActuatorTypeR8_16')
ActuatorType = EnumType('ActuatorType', [ActuatorTypeX5_1, ActuatorTypeX5_4, ActuatorTypeX5_9, ActuatorTypeX8_3, ActuatorTypeX8_9, ActuatorTypeX8_16, ActuatorTypeR8_3, ActuatorTypeR8_9, ActuatorTypeR8_16])

# Link Types
LinkTypeX5 = EnumTraits(0, 'LinkTypeX5')
LinkTypeR8 = EnumTraits(1, 'LinkTypeR8')
LinkType = EnumType('LinkType', [LinkTypeX5, LinkTypeR8])

# Link Input Types
LinkInputTypeRightAngle = EnumTraits(0, 'LinkInputTypeRightAngle')
LinkInputTypeInline = EnumTraits(1, 'LinkInputTypeInline')
LinkInputType = EnumType('LinkInputType', [LinkInputTypeRightAngle, LinkInputTypeInline])

# Link Output Types
LinkOutputTypeRightAngle = EnumTraits(0, 'LinkOutputTypeRightAngle')
LinkOutputTypeInline = EnumTraits(1, 'LinkOutputTypeInline')
LinkOutputType = EnumType('LinkOutputType', [LinkOutputTypeRightAngle, LinkOutputTypeInline])

# Bracket Types
BracketTypeX5LightLeft = EnumTraits(0, 'BracketTypeX5LightLeft')
BracketTypeX5LightRight = EnumTraits(1, 'BracketTypeX5LightRight')
BracketTypeX5HeavyLeftInside = EnumTraits(2, 'BracketTypeX5HeavyLeftInside')
BracketTypeX5HeavyLeftOutside = EnumTraits(3, 'BracketTypeX5HeavyLeftOutside')
BracketTypeX5HeavyRightInside = EnumTraits(4, 'BracketTypeX5HeavyRightInside')
BracketTypeX5HeavyRightOutside = EnumTraits(5, 'BracketTypeX5HeavyRightOutside')
BracketTypeR8LightLeft = EnumTraits(6, 'BracketTypeR8LightLeft')
BracketTypeR8LightRight = EnumTraits(7, 'BracketTypeR8LightRight')
BracketTypeR8HeavyLeftInside = EnumTraits(8, 'BracketTypeR8HeavyLeftInside')
BracketTypeR8HeavyLeftOutside = EnumTraits(9, 'BracketTypeR8HeavyLeftOutside')
BracketTypeR8HeavyRightInside = EnumTraits(10, 'BracketTypeR8HeavyRightInside')
BracketTypeR8HeavyRightOutside = EnumTraits(11, 'BracketTypeR8HeavyRightOutside')
BracketType = EnumType('BracketType', [BracketTypeX5LightLeft, BracketTypeX5LightRight, BracketTypeX5HeavyLeftInside, BracketTypeX5HeavyLeftOutside, BracketTypeX5HeavyRightInside, BracketTypeX5HeavyRightOutside, BracketTypeR8LightLeft, BracketTypeR8LightRight, BracketTypeR8HeavyLeftInside, BracketTypeR8HeavyLeftOutside, BracketTypeR8HeavyRightInside, BracketTypeR8HeavyRightOutside])

EndEffectorTypeCustom = EnumTraits(0, 'EndEffectorTypeCustom')
EndEffectorTypeX5Parallel = EnumTraits(1, 'EndEffectorTypeX5Parallel')
EndEffectorTypeR8Parallel = EnumTraits(2, 'EndEffectorTypeR8Parallel')
EndEffectorType = EnumType('EndEffectorType', [EndEffectorTypeCustom, EndEffectorTypeX5Parallel, EndEffectorTypeR8Parallel])

MatrixOrderingRowMajor = EnumTraits(0, 'MatrixOrderingRowMajor')
MatrixOrderingColumnMajor = EnumTraits(1, 'MatrixOrderingColumnMajor')
MatrixOrdering = EnumType('MatrixOrdering', [MatrixOrderingRowMajor, MatrixOrderingColumnMajor])

# Generated enums for all message field enums
from ._message_enums import *



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------


import numpy as np


def is_finite(m):
  """
  Determine if the input has all finite values.

  Deprecation notice: It is deprecated to pass a `str` in as a parameter here.
  This functionality will be removed in a future release.

  :param m: a matrix or array of any shape and size
  :type m:  str, list, numpy.ndarray, ctypes.Array
  """
  if isinstance(m, str):
    # Deprecated: to be removed in a future version
    m = np.asmatrix(m, dtype=np.float64)
  res = np.isfinite(m)
  if isinstance(res, bool):
    return res
  return res.all()


def is_so3_matrix(m):
  """
  Determine if the matrix belongs to the SO(3) group.
  This is found by calculating the determinant and seeing if it equals 1.

  Deprecation notice: It is deprecated to pass a `str` in as a parameter here.
  This functionality will be removed in a future release.

  :param m: a 3x3 matrix
  :type m:  str, list, numpy.ndarray, ctypes.Array
  """
  try:
    # Use highest possible precision
    if isinstance(m, str):
      # Deprecated: to be removed in a future version
      m = np.asmatrix(m, dtype=np.float64).reshape((3, 3))
    det = np.linalg.det(m)

    # Arbitrarily determined. This may change in the future.
    tolerance = 1e-5
    diff = abs(det-1.0)

    return diff < tolerance
  except Exception as e:
    return False



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018-2019 HEBI Robotics
#  See http://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------
from os import path
from setuptools import setup, Extension

api_version = '2.1.0'
allow_upload = True

if (api_version == '{0}{1}{2}'.format('#', 'PY_', 'VERSION_INSERT')):
  allow_upload = False
  print('Warning: Version number not generated. Not allowing upload to PyPI.')
  import sys
  if 'upload' in sys.argv:
    print('Error: Uploading to PyPI without generating api_version is explicitly disabled. Exiting.')
    exit(1)

api_reference_url = f"http://docs.hebi.us/docs/python/{api_version}"
changelog_url     = "http://docs.hebi.us/downloads_changelogs.html#python-api-changelog"
documentation_url = "http://docs.hebi.us/tools.html#python-api"
license_url = "https://www.hebirobotics.com/softwarelicense"

description = f"""
HEBI Core Python API
====================

HEBI Python provides bindings for the HEBI Core library.

API Reference available at {api_reference_url}

Documentation available on [docs.hebi.us]({documentation_url}).

Refer to the [API changelog]({changelog_url}) for version history.

By using this software, you agree to our [software license]({license_url}).
"""


setup(name                          = 'hebi-py',
      version                       = api_version,
      description                   = 'HEBI Core Python bindings',
      long_description              = description,
      long_description_content_type ="text/markdown",
      author                        = 'Daniel Wright',
      author_email                  = 'support@hebirobotics.com',
      url                           = 'https://docs.hebi.us',
      packages                      = ['hebi'],
      package_data                  = {'hebi': [
        'lib/**/libhebi.so*',
        'lib/**/libhebiWrapper.so*',
        'lib/**/libhebi.*dylib',
        'lib/**/libhebiWrapper.*dylib',
        'lib/**/hebi.dll',
        'lib/**/hebiWrapper.dll',
        '_internal/*',
        '_internal/ffi/*'
      ]},
      install_requires              = [
        'numpy'
      ],
      classifiers                   = [
        "Development Status :: 5 - Production/Stable",
        "Environment :: MacOS X",
        "Intended Audience :: Developers",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: Implementation :: CPython",
      ],
      project_urls                  = {
        "API Reference" : api_reference_url,
        "Changelog"     : changelog_url,
        "Documentation" : documentation_url,
        "Examples"      : "https://github.com/HebiRobotics/HEBI-python-examples"
      }
      )
